ENT.Type		= "brush"
ENT.Base		= "base_cp_entity"
ENT.PrintName		= "base_cp_trigger"
ENT.Author		= "mrflippy"
ENT.Spawnable		= false
ENT.AdminSpawnable	= false

ENT.PercentRequiredToCap = 100;

--list of the participants in the trigger area
InitAccessorFunc(ENT, "Participants", {});

--the percentage capped (0 - PercentRequiredToCap)
InitAccessorFunc(ENT, "CapPercentage", 0);

--how much time has been accrued so far
InitAccessorFunc(ENT, "AccruedTime", 0);

--the last time we changed time
InitAccessorFunc(ENT, "LastTimeChange", 0);

--Id of the team that owns the current time on the clock
InitAccessorFunc(ENT,"CapturingTeamId", nil);

--table to hold data about the teams that are currently capturing
InitAccessorFunc(ENT, "CapturingTeams", {});

--var to hold the number of current cappers
InitAccessorFunc(ENT, "NumberOfCappers", 0);

function ENT:Initialize(thisClass)
	self:BaseClassInitialize(thisClass)

	self:ProcessBaseCPTriggerKeyValues();
end

function ENT:AcceptInput(name, activator, caller)
	print("trigger_cp_base Input: "..name);
end

function ENT:Touch(ent)
end

--determine wheter the entity should be added to area
function ENT:PassesTriggerFilters(ent)
	if(ent:IsPlayer()) then
		return true
	else
		return false;
	end
end

--TODO: Eventually allow things other than players?
function ENT:StartTouch(ent)
	if(!self:PassesTriggerFilters(ent)) then return end

	print(self:GetName()..": StartTouch");

	self:AddParticipant(ent);
end

function ENT:EndTouch(ent)
	if(!self:PassesTriggerFilters(ent)) then return end

	print(self:GetName()..": EndTouch");
	self:RemoveParticipant(ent);
end

function ENT:Think()
	--figure out who can capture and block
	--call this every think in case things change
	self:ProcessParticipants();

	local capTeamCount = table.Count(self:GetCapturingTeams());

	if(capTeamCount == 0) then
		--check for no cappers
		self:DecreaseTime();

		if(self:GetCapturingTeamId() != nil and self:GetCapPercentage() <= 0) then
			--reset the capturing team id if the time is gone
			print("Resetting Time");
			self:SetCapturingTeamId(nil);

			--call the expired capture event
			hook.Call("CP_OnExpireCap", self.CapturePoint, self);
		end
	elseif(capTeamCount == 1) then
		--if there's only 1 team then that team starts capturing or reverting
		--
		--pull the team id
		local teamId = next(self:GetCapturingTeams());
		
		--print(self:GetName(),": Participants: ", table.Count(self.Participants));

		--save the previous NumberOfCappers
		local previousNumberOfCappers = self:GetNumberOfCappers();
	
		self:SetNumberOfCappers(table.Count(self:GetCapturingTeams()[teamId]));

		--if(self:GetNumberOfCappers() != previousNumberOfCappers) then
			--send OnNumberOfCappersChanged output
			--TODO: valve wiki says "Sent when the number of people in the capture area changes"
			--Is this correct?
		--end
	
		print("CapturingTeamId: ", self:GetCapturingTeamId(), ", OwningTeamId: ", self.CapturePoint:GetOwningTeamId());

		--TODO: this logic needs to be changed if we want to support multiple capture areas
		--In order to increase time:
		--team needs to be able to capture (cannot be defense team or already own the point) (handled in ProcessParticipants)
		--if existing time is on the clock, it must belong to the team
		if ((self:GetCapturingTeamId() == nil or self:GetCapturingTeams()[self:GetCapturingTeamId()])
			and self.CapturePoint:GetOwningTeamId() != teamId) then

			if(self:GetCapturingTeamId() == nil or self.CaptureWasBlocked) then
				--raise the CP_OnStartCapture event
				hook.Call("CP_OnStartCap", self.CapturePoint, self);

				--sent the OnStartCap output
				self:TriggerOutput("OnStartCap");

				--send the OnStartTeam<teamId> output
				--TODO: subtraction is for TF2! Need some sort of mapping function
				--self:TriggerOutput("OnStartTeam"..(teamId - 1));
				self:TriggerOutput("OnStartTeam"..teamId);
			end

			--reset the WasBlocked flag
			self.CaptureWasBlocked = false;
			
			--set the capturing team id
			print(self.PrintName, ": Setting capturing team id to: ", teamId)
			self:SetCapturingTeamId(teamId);

			--start capturing
			self:IncreaseTime();

			print("AccruedTime: ", self:GetAccruedTime(), "area_time_to_cap: ", self:GetCPKeyValues()["area_time_to_cap"], "CapPercentage:", self:GetCapPercentage())
			if(self:GetCapPercentage() >= self.PercentRequiredToCap) then
				--captured!
				--raise the CP_OnPointCaptured event
				hook.Call("CP_OnEndCap", self.CapturePoint, self);

				--send OnCapTeam<teamId> output
				--TODO: the subtraction is for TF2! how to handle?
				--self:TriggerOutput("OnCapTeam"..(teamId - 1));
				self:TriggerOutput("OnCapTeam"..teamId);

				--send OnEndCap output
				self:TriggerOutput("OnEndCap");

				--reset the vars
				--TODO: this should probably go elsewhere
				self:SetCapturingTeamId(nil);
				self:SetAccruedTime(0);
				self:SetLastTimeChange(0);
			end
		else
			--reverting capture
			--(capture blocked by time accrued by other team
			--or the team is the defending team)
			self:DecreaseTime();
		end
	else
		--check for new block
		if(!self.CaptureWasBlocked) then
			self.CaptureWasBlocked = true;

			--raise the CP_OnBreakCap event
			--TODO: easier way for event handlers to get blocking teams?
			hook.Call("CP_OnBreakCap", self.CapturePoint, self);

			--send OnBreakTeam output
			--TODO: subtraction is for TF2
		--	self:TriggerOutput("OnBreakTeam.."..(teamId - 1));
			self:TriggerOutput("OnBreakTeam.."..teamId);

			--send OnBreakCap output
			self:TriggerOutput("OnBreakCap");
		end
		
		--capture blocked by entities
		--no time decrease if both capture and defense
	end
end

--Increases the time on the clock by the difference between now
--and the previous call of this function
function ENT:IncreaseTime()
	local newTime = CurTime();

	--since we're working with time differences and CurTime()
	--returns an elapsed time from a start point, a time of 0
	--will screw things up
	if(self:GetLastTimeChange() == 0) then
		self:SetLastTimeChange(newTime);
	end

	--accrue the elapsed time since the last time this was called
	self:SetAccruedTime(self:GetAccruedTime() + (newTime - self:GetLastTimeChange()));

	--set the new time
	self:SetLastTimeChange(newTime);

	self:CalculateCapPercentage();
end

--Decreases the time on the clock by the difference between now
--and the previous call of this function
function ENT:DecreaseTime()
	--don't decrease time below 0
	if(self:GetAccruedTime() <= 0) then
		self:SetAccruedTime(0);
		self:SetLastTimeChange(0);

		return
	end

	local newTime = CurTime();

	--since we're working with time differences and CurTime()
	--returns an elapsed time from a start point, a time of 0
	--will screw things up
	if(self:GetLastTimeChange() == 0) then
		self:SetLastTimeChange(newTime);
	end

	--don't allow the timer to go below 0
	local newAccruedTime = self:GetAccruedTime() - (newTime - self:GetLastTimeChange());
	if(newAccruedTime < 0) then newAccruedTime = 0 end

	--accrue the elapsed time since the last time this was called
	self:SetAccruedTime(newAccruedTime);

	--set the new time
	self:SetLastTimeChange(newTime);

	print("Decreasing time to: ", newAccruedTime);

	self:CalculateCapPercentage();
end

--calcuates the percentage based on the accrued time and area_time_to_cap
--TODO: area_time_to_cap should be passed through a filter first
function ENT:CalculateCapPercentage()
	local percentage = (self:GetAccruedTime()/tonumber(self:GetCPKeyValues()["area_time_to_cap"])) * 100;

	if(percentage > self.PercentRequiredToCap) then
		percentage = self.PercentRequiredToCap;
	else
		percentage = math.Round(percentage);
	end

	self:SetCapPercentage(percentage);
end

--categorizes the participants into the CapturingTeams table
function ENT:ProcessParticipants()
	--clear out the CapturingTeams table
	self:SetCapturingTeams({});

	--loop through the participants
	for entId, ent in pairs(self:GetParticipants()) do
		--save the team
		local participantTeam = self:GetTeam(ent);

		--check whether the team and participants are allowed to capture
		if(self:CanTeamCapture(participantTeam) and self:CanParticipantCapture(ent)) then
			if(self:GetCapturingTeams()[participantTeam] == nil) then
				self:GetCapturingTeams()[participantTeam] = {};
			end

			--if we get here, add the participant to the capturing teams table
			self:GetCapturingTeams()[participantTeam][entId] = ent;
		end
	end
end

--Adds an entity to the participant list
function ENT:AddParticipant(entityToAdd)
	if(!entityToAdd:IsValid()) then return end

	--adds the entity
	self:GetParticipants()[entityToAdd:EntIndex()] = entityToAdd;

	--send OnNumCappersChanged output
	self:TriggerOutput("OnNumCappersChanged");

	--raise CP_ParticipantAdded event
	hook.Call("CP_ParticipantAdded", nil, self, entityToAdd);
end

--Removes and entity from the participant list
function ENT:RemoveParticipant(entityToRemove)
	if(!entityToRemove:IsValid()) then return end

	--removes the entity from the list
	self:GetParticipants()[entityToRemove:EntIndex()] = nil;

	--send OnNumCappersChanged output
	self:TriggerOutput("OnNumCappersChanged");

	--raise CP_ParticipantRemoved event
	hook.Call("CP_ParticipantRemoved", nil, self, entityToRemove);
end

--determine whether a team is allowed to capture
--TODO: check key/values
--TODO: implement callback?
function ENT:CanTeamCapture(teamId)
	--return true for now
	return true;
end

--determine whether a participant is allowed to capture
function ENT:CanParticipantCapture(ent)
	--return true for now
	if(ent:IsPlayer()) then
		return true;
	else
		return false;
	end
end

--determine the entity's team
--TODO: implement callback
function ENT:GetTeam(ent)
	if(!ent:IsValid()) then return end

	--TODO: change to check for ent.Team
	if(ent:IsPlayer()) then
		return ent:Team();
	else
		return 1;
	end
end

--processes this entity's key values
--TODO: callbacks?
function ENT:ProcessBaseCPTriggerKeyValues()
	local area_cap_point_name = self:GetCPKeyValues()["area_cap_point"];
	if(area_cap_point_name) then
		--the cap point this area is linked to
		--TODO: error checking
		self.CapturePoint = ents.FindByName(area_cap_point_name)[1];
	end
end
